MOTIF_ICONS = {
    # Connection motifs
    "Tetraloop": "bi bi-braces",
    "single_strand": "bi bi-dash-square",
    "start_end_stem": "bi bi-skip-end-fill",
    "stem_cap_link": "bi bi-align-start",
    "vertical_double_link": "bi bi-arrow-down-up",
    "vertical_link": "bi bi-arrow-up",
    # Structural motifs
    "Stem": "bi bi-layers",
    "Dovetail": "bi bi-fullscreen-exit",
    "LambdaTurn": "bi bi-sign-turn-slight-right",
    "ThreeWayJunction": "bi bi-sign-intersection-side",
    "Bend90": "bi bi-arrow-90deg-right",
    # Kissing loop motifs
    "KissingDimer": "bi bi-yin-yang",
    "KissingLoop180": "bi bi-square-half",
    "KissingLoop120": "bi bi-triangle-half",
    "BranchedKissingLoop": "bi bi-align-top",
    "BranchedDimer": "bi bi-tree",
    "KissingDimer120": "bi bi-send",
    # Common aptamers
    "Broccoli": "bi bi-lightbulb",
    "Pepper": "bi bi-stoplights",
    "Ispinach": "bi bi-lamp",
    "MS2": "bi bi-signpost",
    "Streptavidin": "bi bi-link-45deg",
    None: "bi bi-bezier2",
}
